from __future__ import print_function
from problog.logic import Term, Var, Clause, And, Constant
from problog.engine_builtin import StructSort
from problog.util import OrderedSet

class Rule(object):
    """Generic class for rules."""

    def __init__(self):
        pass

    def get_literals(self):
        """Returns all literals in the rule."""
        raise NotImplementedError('abstract method')

    def get_literal(self):
        """Returns the last added literal."""
        raise NotImplementedError('abstract method')

    def get_variables(self):
        """Return the set of variables in the rule.

        :return: all variables in that occur in the rule
        :rtype: set[Var]
        """
        variables = OrderedSet()
        for lit in self.get_literals():
            variables |= lit.variables()
        return variables

    def to_clauses(self, functor=None):
        """Transform rule into ProbLog clauses

        :param functor: override rule functor (set to None to keep original)
        :type functor: str | None
        :return: clause representation
        :rtype: list[Clause]
        """
        raise NotImplementedError('abstract method')


class FOILRuleB(Rule):
    """A FOIL rule is a rule with a specified target literal.
    These rules are represented as a (reversed) linked list.

    :param parent: parent rule which corresponds to the rule obtained by removing the last literal
    :type parent: FOILRuleB | None
    :param literal: last literal of the body
    :type literal: Term | None
    """
    def __init__(self, parent, literal, target, previous, correct):
        Rule.__init__(self)
        self.parent = parent
        self.literal = literal
        self.target = target
        self.previous = previous
        self.correct = correct
        self.rule_prob = None
        self.score = None

    def get_literals(self):
        """Get literals in the rule.

        :return: list of literals including target
        :rtype: list[Term]
        """
        return self.parent.get_literals() + [self.literal]

    def get_literal(self):
        """Get most recently added body literal.

        :return: None (the body is empty for this type of rule)
        :rtype: Term
        """
        return self.literal

    def __and__(self, literal):
        """Add a literal to the body of the rule.

        :param literal: literal to add
        :type literal: Term
        :return: new rule
        :rtype: FOILRuleB
        """
        return FOILRuleB(self, literal, self.target, self.previous, self.correct)

    def to_clauses(self, functor=None):
        """Transform rule into ProbLog clauses

        :param functor: override rule functor (set to None to keep original)
        :type functor: str | None
        :return: clause representation
        :rtype: list[Clause]
        """
        if self.previous:
            previous = self.previous.to_clauses(functor)
        else:
            previous = []

        inequalities = []

        object_identity = False
        if object_identity:
            variables = list(self.get_variables())
            for i, v1 in enumerate(variables):
                for v2 in variables[i + 1:]:
                    inequalities.append(Term('\=', v1, v2))

        literals = self.get_literals()

        def rename_recursive(lit, functor):
            if lit.functor == '_recursive':
                return Term(functor, *lit.args)
            else:
                return lit

        literals = [rename_recursive(lit, functor) for lit in literals]

        prob = self.get_rule_probability()
        if prob is not None:
            prob = Constant(prob)
        head = literals[0].with_probability(prob)
        body = And.from_list(literals[1:] + inequalities)

        if functor is not None:
            head = Term(functor, *head.args, p=prob)
        return previous + [Clause(head, body)]

    def set_rule_probability(self, probability=None):
        rule = self
        while rule.parent:
            rule = rule.parent
        rule.rule_prob = probability

    def get_rule_probability(self):
        rule = self
        while rule.parent:
            rule = rule.parent
        return rule.rule_prob

    def has_new_variables(self):
        if not self.parent:
            return self.target.variables() - self.parent.get_variables()
        else:
            return self.get_literal().variables() - self.parent.get_variables()

    def __str__(self):
        literals = self.get_literals()
        head = literals[0].with_probability(self.get_rule_probability())
        if len(literals) == 1:
            return '%s :- true.' % (head,)
        else:
            return '%s :- %s' % (head, ', '.join(map(str, literals[1:])))

    def is_equivalent(self, other):
        if abs(self.score - other.score) > 1e-8:
            return False
        literals1 = sorted(self.get_literals(), key=StructSort)
        literals2 = sorted(other.get_literals(), key=StructSort)

        result = literals1 == literals2
        return result

    def __len__(self):
        if self.parent:
            return len(self.parent) + 1
        else:
            return 1


class FOILRule(FOILRuleB):
    """Represents the head of a FOILRule.

    :param target: literal in the head of the rule
    :type target: Term
    :param previous: previous rule (if part of set)
    :type previous: FOILRuleB
    """

    def __init__(self, target=None, previous=None, correct=None):
        if target is None and previous is not None:
            target = previous.target
        FOILRuleB.__init__(self, None, None, target, previous, correct)

    def get_literals(self):
        """Get literals in the rule.

        :return: list of literals including target
        :rtype: list[Term]
        """
        return [self.target]

    def get_literal(self):
        """Get most recently added body literal.

        :return: None (the body is empty for this type of rule)
        :rtype: Term
        """
        return None

    def __str__(self):
        return '%s :- true' % self.target


from __future__ import print_function


def main(argv=sys.argv[1:]):
    args = argparser().parse_args(argv)

def argparser():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('model')
    parser.add_argument('data')
    return parser

if __name__ == '__main__':
    main()



from __future__ import print_function


from collections import defaultdict
from itertools import product

from problog.logic import Term, is_variable, Var, Clause, And


class BaseLanguage(object):
    """Base class for languages."""

    def __init__(self):
        pass

    def refine(self, rule):
        """Generate one refinement for the given rule.

        :param rule: rule for which to generate refinements
        """
        raise NotImplementedError('abstract method')


class TypeModeLanguage(BaseLanguage):
    """Typed and Mode-based language."""

    MODE_EXIST = '+'     # use existing variable
    MODE_NEW = '-'       # create new variable (do not reuse existing)
    MODE_CONSTANT = 'c'  # insert constant

    def __init__(self, symmetry_breaking=True, **kwargs):
        BaseLanguage.__init__(self)

        self._types = {}    # dict[tuple[str,int],tuple[str]]: signature / argument types
        self._values = defaultdict(set)   # dict[str, set[Term]]: values in data for given type
        self._modes = []    # list[tuple] : list of functor, modestr pairs

        self._symmetry_breaking = symmetry_breaking
        self._allow_negation = True
        self._allow_recursion = False

    def add_types(self, functor, argtypes):
        """Add type information for a predicate.

        Type information has to be unique for a functor/arity combination.

        :param functor: functor of the predicate
        :type functor: str
        :param argtypes: types of the arguments (arity is length of this list)
        :type argtypes: list[str]
        :raise ValueError: duplicate type definition is given
        """
        key = (functor, len(argtypes))
        if key in self._types:
            raise ValueError("A type definition already exists for '%s/%s'."
                             % (functor, len(argtypes)))
        else:
            self._types[key] = argtypes

    def add_modes(self, functor, argmodes):
        """Add mode information for a predicate.

        :param functor: functor of the predicate
        :type functor: str
        :param argmodes: modes of the arguments (arity is length of this list)
        :type argmodes: list[str]
        """
        self._modes.append((functor, argmodes))

    def add_values(self, typename, *values):
        """Add a value for a predicate.

        :param typename: name of the type
        :type typename: str
        :param values: value to add (can be multiple)
        :type values: collections.Iterable[Term]
        """
        for value in values:
            self._values[typename].add(value)

    def refine(self, rule):
        """Generate one refinement for the given rule.

        :param rule: rule for which to generate refinements
        :return: generator of literals that can be added to the rule
        :rtype: collections.Iterable[Term]
        """
        varcount = len(rule.get_variables())
        variables = self.get_variable_types(*rule.get_literals())

        if rule.get_literal():
            if rule.get_literal().functor == '_recursive':
                return  # can't extend after recursive
            generated = rule.get_literal().refine_state
        else:
            generated = set()

        # 1) a positive refinement
        for functor, modestr in self._modes:
            arguments = []
            arity = len(modestr)
            types = self.get_argument_types(functor, arity)
            for argmode, argtype in zip(modestr, types):
                if argmode == '+':
                    # All possible variables of the given type
                    arguments.append(variables.get(argtype, []))
                elif argmode == '-':
                    # A new variable
                    arguments.append([Var('#')])  # what about adding a term a(X,X) where X is new?
                elif argmode == 'c':
                    # Add a constant
                    arguments.append(self.get_type_values(argtype))
                    pass
                else:
                    raise ValueError("Unknown mode specifier '%s'" % argmode)
            for args in product(*arguments):
                t = Term(functor, *args)
                if t not in generated:
                    if self._symmetry_breaking:
                        generated.add(t)
                    t_i = t.apply(TypeModeLanguage.ReplaceNew(varcount))
                    t_i.prototype = t
                    if self._symmetry_breaking:
                        t_i.refine_state = generated.copy()
                    else:
                        t_i.refine_state = generated | {t, -t, t_i, -t_i}
                    yield t_i

        # 2) a negative refinement
        if self._allow_negation:
            for functor, modestr in self._modes:
                if '-' in modestr:
                    # No new variables allowed for negative literals
                    continue
                arguments = []
                arity = len(modestr)
                types = self.get_argument_types(functor, arity)
                for argmode, argtype in zip(modestr, types):
                    if argmode == '+':
                        # All possible variables of the given type
                        arguments.append(variables.get(argtype, []))
                    elif argmode == 'c':
                        # Add a constant
                        arguments.append(self.get_type_values(argtype))
                        pass
                    else:
                        raise ValueError("Unknown mode specifier '%s'" % argmode)
                for args in product(*arguments):
                    t = -Term(functor, *args)
                    if t not in generated:
                        if self._symmetry_breaking:
                            generated.add(t)
                        t_i = t.apply(TypeModeLanguage.ReplaceNew(varcount))
                        t_i.prototype = t
                        if self._symmetry_breaking:
                            t_i.refine_state = generated.copy()
                        else:
                            t_i.refine_state = generated | {t, -t, t_i, -t_i}
                        yield t_i

        # 3) recursive
        if self._allow_recursion and rule.previous.previous:  # recursion in the first rule makes no sense
            types = self.get_argument_types(rule.target.functor, rule.target.arity)

            arguments = []
            for argtype in types:
                arguments.append(variables.get(argtype, []))

            for args in product(*arguments):
                t = Term('_recursive', *args)
                t.prototype = t
                if self._symmetry_breaking:
                    generated.add(t)
                if self._symmetry_breaking:
                    t.refine_state = generated.copy()
                else:
                    t.refine_state = generated | {t}
                yield t

    def get_type_values(self, typename):
        """Get all values that occur in the data for a given type.

        :param typename: name of type
        :type typename: str
        :return: set of values
        :rtype: set[Term]
        """
        return self._values.get(typename, [])

    def get_argument_types(self, functor, arity):
        """Get the types of the arguments of the given predicate.

        :param functor: functor of the predicate
        :type functor: str
        :param arity: arity of the predicate
        :type arity: int
        :return: tuple of type descriptors, one for each argument
        :rtype: tuple[str]
        """
        return self._types[(functor, arity)]

    def get_variable_types(self, *literals):
        """Get the types of all variables that occur in the given literals.

        :param literals: literals to extract variables from
        :type literals: collections.Iterable[Term]
        :return: dictionary with list of variables for each type
        :rtype: dict[str, list[Term]]
        """
        result = defaultdict(set)
        for lit in literals:
            if lit.is_negated():
                lit = -lit

            if lit.functor == '_recursive':
                # Don't need to process this, variables will occur somewhere else
                #  because _recursive has mode + on all arguments.
                continue
            types = self.get_argument_types(lit.functor, lit.arity)
            for arg, argtype in zip(lit.args, types):
                if is_variable(arg) or arg.is_var():
                    result[argtype].add(arg)
        return result

    def load(self, data):
        """Load from data.

        :param data: datafile
        :type data: DataFile
        """

        for typeinfo in data.query('base', 1):
            typeinfo = typeinfo[0]
            self.add_types(typeinfo.functor, list(map(str, typeinfo.args)))

        for modeinfo in data.query('mode', 1):
            modeinfo = modeinfo[0]
            self.add_modes(modeinfo.functor, list(map(str, modeinfo.args)))

        for predicate, types in self._types.items():
            arg_values = zip(*data.query(*predicate))
            for a, t in zip(arg_values, types):
                self.add_values(t, *a)

        for optname, optvalue in data.query('option', 2):
            if str(optname) == 'negation' and str(optvalue) == 'off':
                self._allow_negation = False
            elif str(optname) == 'recursion' and str(optvalue) == 'on':
                self._allow_recursion = True

    class ReplaceNew(object):
        """Helper class for replacing new variables (indicated by name '#') by unique variables.

        :param count: the current number of variables
        :type count: int
        """

        def __init__(self, count):
            self.count = count

        def _get_name(self, index):
            """Get a name for the new variable.
            This name will be a single letter from A to Z, or V<num> if index > 25.

            :param index: number of the variable
            :type index: int
            :return: name of variable
            :rtype: str
            """
            if index < 26:
                return chr(65 + index)
            else:
                return 'V%d' % index

        def __getitem__(self, name):
            if name == '#':
                name = self._get_name(self.count)
                self.count += 1
                return Var(name)
            else:
                return Var(name)



"""Implementation of the Prob2FOIL algorithm.


"""

from __future__ import print_function

from problog.program import PrologFile
from .data import DataFile
from .language import TypeModeLanguage
from problog.util import init_logger
from problog.logic import Term
from .rule import FOILRule
from .learn import CandidateBeam, LearnEntail

from logging import getLogger

import time
import argparse
import sys
import random

from .score import rates, accuracy, m_estimate_relative, precision, recall, m_estimate_future_relative, significance, pvalue2chisquare


class ProbFOIL(LearnEntail):

    def __init__(self, data, beam_size=5, logger='probfoil', m=1, l=None, p=None, **kwargs):
        LearnEntail.__init__(self, data, TypeModeLanguage(**kwargs), logger=logger, **kwargs)

        self._beamsize = beam_size
        self._m_estimate = m
        if p is None:
            self._min_significance = None
        else:
            self._min_significance = pvalue2chisquare(p)

        self.load(data)   # for types and modes

        getLogger(self._logger).info('Number of examples (M): %d' % len(self._examples))
        getLogger(self._logger).info('Positive weight (P): %.4f' % sum(self._scores_correct))
        getLogger(self._logger).info('Negative weight (N): %.4f' %
                                     (len(self._scores_correct) - sum(self._scores_correct)))

        self.interrupted = False
        self._stats_evaluations = 0
        self._max_length = l

    def best_rule(self, current):
        """Find the best rule to extend the current rule set.

        :param current:
        :return:
        """

        current_rule = FOILRule(target=self.target, previous=current, correct=self._scores_correct)
        current_rule.scores = [1.0] * len(self._scores_correct)
        current_rule.score = self._compute_rule_score(current_rule)
        c_tp, c_fp, c_tn, c_fn = rates(current_rule)
        current_rule.score_cmp = (current_rule.score, c_tp)
        current_rule.processed = False
        current_rule.probation = False
        current_rule.avoid_literals = set()
        if current:
            prev_tp = rates(current)[0]
        else:
            prev_tp = 0.0

        best_rule = current_rule

        self.interrupted = False
        try:
            candidates = CandidateBeam(self._beamsize)
            candidates.push(current_rule)
            iteration = 1
            while candidates:
                next_candidates = CandidateBeam(self._beamsize)

                getLogger(self._logger).debug('Best rule so far: %s [%s]' % (best_rule, best_rule.score))
                getLogger(self._logger).debug('Candidates for iteration %s:' % iteration)
                getLogger(self._logger).debug(candidates)
                iteration += 1
                while candidates:
                    current_rule = candidates.pop()
                    current_rule_literal_avoid = set(current_rule.avoid_literals)
                    getLogger(self._logger).log(8, 'TO AVOID: %s => %s' % (current_rule, current_rule.avoid_literals))
                    c_tp, c_fp, c_tn, c_fn = rates(current_rule)
                    if self._max_length and len(current_rule) >= self._max_length:
                        pass
                    else:
                        for ref in self.language.refine(current_rule):
                            if ref in current_rule.avoid_literals:  # or ref.prototype in current_rule.avoid_literals:
                                getLogger(self._logger).log(8, 'SKIPPED literal %s for rule %s' % (ref, current_rule))
                                continue
                            rule = current_rule & ref
                            getLogger(self._logger).log(7, 'EVALUATING RULE %s' % rule)
                            rule.scores = self._compute_scores_predict(rule)
                            self._stats_evaluations += 1
                            rule.score = self._compute_rule_score(rule)
                            r_tp, r_fp, r_tn, r_fn = rates(rule)
                            rule.score_cmp = (rule.score, r_tp)
                            rule.score_future = self._compute_rule_future_score(rule)
                            rule.processed = False
                            rule.avoid_literals = current_rule_literal_avoid

                            if prev_tp > r_tp - 1e-8:       # new rule has no tp improvement over previous
                                getLogger(self._logger).log(8, '%s %s %s %s [REJECT coverage] %s' % (rule, rule.score, rates(rule), rule.score_future, prev_tp))
                                # remove this literal for all sibling rules
                                current_rule_literal_avoid.add(ref)
                                current_rule_literal_avoid.add(ref.prototype)
                            elif rule.score_future <= best_rule.score:
                                getLogger(self._logger).log(8, '%s %s %s %s [REJECT potential] %s' % (rule, rule.score, rates(rule), rule.score_future, best_rule.score))
                                # remove this literal for all sibling rules
                                current_rule_literal_avoid.add(ref)
                                current_rule_literal_avoid.add(ref.prototype)
                            elif r_fp > c_fp - 1e-8:  # and not rule.has_new_variables():
                                # no fp eliminated and no new variables
                                getLogger(self._logger).log(8, '%s %s %s %s [REJECT noimprov] %s' % (rule, rule.score, rates(rule), rule.score_future, best_rule.score))
                                # remove this literal for all sibling rules
                                # current_rule_literal_avoid.add(ref)
                                # current_rule_literal_avoid.add(ref.prototype)
                            elif r_fp > c_fp - 1e-8 and current_rule.probation:
                                getLogger(self._logger).log(8, '%s %s %s %s [REJECT probation] %s' % (rule, rule.score, rates(rule), rule.score_future, best_rule.score))
                            elif r_fp < 1e-8:
                                # This rule can not be improved by adding a literal.
                                # We reject it for future exploration,
                                #  but we do consider it for best rule.
                                getLogger(self._logger).log(9, '%s %s %s %s [REJECT* fp] %s' % (rule, rule.score, rates(rule), rule.score_future, prev_tp))
                                if rule.score_cmp > best_rule.score_cmp:
                                    getLogger(self._logger).log(9, 'BETTER RULE %s %s > %s' % (rule, rule.score_cmp, best_rule.score_cmp))
                                    best_rule = rule
                            else:
                                if r_fp > c_fp - 1e-8:
                                    rule.probation = True
                                else:
                                    rule.probation = False
                                if next_candidates.push(rule):
                                    getLogger(self._logger).log(9, '%s %s %s %s [ACCEPT]' % (rule, rule.score, rates(rule), rule.score_future))
                                else:
                                    getLogger(self._logger).log(8, '%s %s %s %s [REJECT beam]' % (rule, rule.score, rates(rule), rule.score_future))

                                if rule.score_cmp > best_rule.score_cmp:
                                    getLogger(self._logger).log(9, 'BETTER RULE %s %s > %s' % (rule, rule.score_cmp, best_rule.score_cmp))
                                    best_rule = rule
                candidates = next_candidates
        except KeyboardInterrupt:
            self.interrupted = True
            getLogger(self._logger).info('LEARNING INTERRUPTED BY USER')

        while best_rule.parent and best_rule.parent.score > best_rule.score - 1e-8:
            best_rule = best_rule.parent

        self._select_rule(best_rule)
        return best_rule

    def initial_hypothesis(self):
        initial = FOILRule(self.target, correct=self._scores_correct)
        initial = initial & Term('fail')
        initial.scores = [0.0] * len(self._scores_correct)
        initial.score = self._compute_rule_score(initial)
        initial.avoid_literals = set()
        return initial

    def learn(self):
        hypothesis = self.initial_hypothesis()
        current_score = 0.0

        while True:
            next_hypothesis = self.best_rule(hypothesis)
            new_score = accuracy(next_hypothesis)
            getLogger(self._logger).info('RULE LEARNED: %s %s' % (next_hypothesis, new_score))

            s = significance(next_hypothesis)
            if self._min_significance is not None and s < self._min_significance:
                break
            if new_score > current_score:
                hypothesis = next_hypothesis
                current_score = new_score
            else:
                break
            if self.interrupted:
                break
            if hypothesis.get_literal() and hypothesis.get_literal().functor == '_recursive':
                break   # can't extend after recursive

        return hypothesis

    def _compute_rule_score(self, rule):
        return m_estimate_relative(rule, self._m_estimate)

    def _compute_rule_future_score(self, rule):
        return m_estimate_future_relative(rule, self._m_estimate)

    def _select_rule(self, rule):
        pass

    def statistics(self):
        return [('Rule evaluations', self._stats_evaluations)]


class ProbFOIL2(ProbFOIL):

    def __init__(self, *args, **kwargs):
        ProbFOIL.__init__(self, *args, **kwargs)

    def _select_rule(self, rule):
        # set rule probability and update scores
        if hasattr(rule, 'max_x'):
            x = round(rule.max_x, 8)
        else:
            x = 1.0

        if x > 1 - 1e-8:
            rule.set_rule_probability(None)
        else:
            rule.set_rule_probability(x)
        if rule.previous is None:
            scores_previous = [0.0] * len(rule.scores)
        else:
            scores_previous = rule.previous.scores

        for i, lu in enumerate(zip(scores_previous, rule.scores)):
            l, u = lu
            s = u - l
            rule.scores[i] = l + x * s

    def _compute_rule_future_score(self, rule):
        return self._compute_rule_score(rule, future=True)

    def _compute_rule_score(self, rule, future=False):
        return self._compute_rule_score_slow(rule, future)

    def _compute_rule_score_slow(self, rule, future=False):
        if rule.previous is None:
            scores_previous = [0.0] * len(rule.scores)
        else:
            scores_previous = rule.previous.scores

        data = list(zip(rule.correct, scores_previous, rule.scores))

        max_x = 0.0
        max_score = 0.0
        max_tp = 0.0
        max_fp = 0.0

        def eval_x(x, data, future=False):
            pos = 0.0
            all = 0.0
            tp = 0.0
            fp = 0.0
            tp_p = 0.0
            fp_p = 0.0
            for p, l, u in data:
                pr = l + x * (u - l)
                tp += min(p, pr)
                fp += max(0, pr - p)
                tp_p += min(p, l)
                fp_p += max(0, l - p)
                pos += p
                all += 1

            if future:
                fp = fp_p
            m = self._m_estimate
            if pos - tp_p == 0 and all - tp_p - fp_p == 0:
                mpnp = 1
            else:
                mpnp = m * ((pos - tp_p) / (all - tp_p - fp_p))
            score = (tp - tp_p + mpnp) / (tp + fp - tp_p - fp_p + m)
            return tp, fp, score

        tp_x, fp_x, score_x = eval_x(1.0, data, future)
        if score_x > max_score:
            max_x = 1.0
            max_tp = tp_x
            max_fp = fp_x
            max_score = score_x
            if not future:
                getLogger('probfoil').log(6, '%s: x=%s (%s %s) -> %s' % (rule, 1.0, tp_x, fp_x, score_x))

        for p, l, u in data:
            if u - l < 1e-8:
                continue
            x = (p - l) / (u - l)

            if x >= 1.0 or x < 0.0:
                continue

            tp_x, fp_x, score_x = eval_x(x, data, future)
            if not future:
                getLogger('probfoil').log(6,
                                          '%s: x=%s (%s %s %s) (%s %s) -> %s' % (rule, x, p, l, u, tp_x, fp_x, score_x))
            if score_x > max_score:
                max_x = x
                max_tp = tp_x
                max_fp = fp_x
                max_score = score_x

        if not future:
            getLogger('probfoil').log(6, '%s: [BEST] x=%s (%s %s) -> %s' % (rule, max_x, max_tp, max_fp, max_score))
            rule.max_x = max_x
            rule.max_tp = max_tp
            rule.max_fp = max_fp

        if max_x < 1e-8:
            return 0.0

        return max_score

    def _compute_rule_score_fast(self, rule, future=False):
        if rule.previous is None:
            scores_previous = [0.0] * len(rule.scores)
        else:
            scores_previous = rule.previous.scores

        pos = 0.0
        all = 0.0

        tp_prev = 0.0
        fp_prev = 0.0
        fp_base = 0.0
        tp_base = 0.0
        ds_total = 0.0
        pl_total = 0.0

        if not future:
            getLogger('probfoil').log(5, '%s: %s' % (rule, list(zip(rule.correct, scores_previous, rule.scores))))

        values = []
        for p, l, u in zip(rule.correct, scores_previous, rule.scores):
            pos += p
            all += 1.0

            tp_prev += min(l, p)
            fp_prev += max(0, l - p)

            ds = u - l  # improvement on previous prediction (note: ds >= 0)
            if ds == 0:  # no improvement
                pass
            elif p < l:  # lower is overestimate
                fp_base += ds
            elif p > u:  # upper is underestimate
                tp_base += ds
            else:   # correct value still achievable
                ds_total += ds
                pl_total += p - l
                y = (p - l) / (u - l)   # for x equal to this value, prediction == correct
                values.append((y, p, l, u))

        neg = all - pos
        mpnp = self._m_estimate * (pos / all)

        def comp_m_estimate(tp, fp):
            score = (tp + mpnp) / (tp + fp + self._m_estimate)
            # print (self._m_estimate, mpnp, tp, fp, score)
            return score

        max_x = 1.0
        tp_x = pl_total + tp_base + tp_prev
        if future:
            fp_x = fp_prev + fp_base
        else:
            fp_x = ds_total - pl_total + fp_base + fp_prev
        score_x = comp_m_estimate(tp_x, fp_x)
        max_score = score_x
        max_tp = tp_x
        max_fp = fp_x

        if values:
            values = sorted(values)
            if not future:
                getLogger('probfoil').log(5, '%s: %s' % (rule, [map(lambda vv: round(vv, 3), vvv) for vvv in values]))

            tp_x, fp_x, tn_x, fn_x = 0.0, 0.0, 0.0, 0.0
            ds_running = 0.0
            pl_running = 0.0
            prev_y = None
            for y, p, l, u in values + [(None, 0.0, 0.0, 0.0)]:     # extra element forces compute at end
                if y is None or prev_y is not None and y > prev_y:
                    # There is a change in y-value.
                    x = prev_y  # set current value of x
                    tp_x = pl_running + x * (ds_total - ds_running) + x * tp_base + tp_prev
                    if future:
                        fp_x = fp_prev
                    else:
                        fp_x = x * ds_running - pl_running + x * fp_base + fp_prev

                    score_x = comp_m_estimate(tp_x, fp_x)

                    if not future:
                        getLogger('probfoil').log(6, '%s: x=%s (%s %s) -> %s' % (rule, x, tp_x, fp_x, score_x))
                    if max_score is None or score_x > max_score:
                        max_score = score_x
                        max_x = x
                        max_tp = tp_x
                        max_fp = fp_x

                        # if not future:
                        #     rts = rates(rule)
                        #     est = m_estimate(rule)
                        #     print(x, tp_x, fp_x, rts, score_x, est)
                        #     # assert abs(tp_x - rts[0]) < 1e-8
                        #     # assert abs(fp_x - rts[1]) < 1e-8
                        #     # assert abs(est - score_x) < 1e-8

                prev_y = y
                pl_running += p - l
                ds_running += u - l

            assert abs(ds_running - ds_total) < 1e-8
            assert abs(pl_running - pl_total) < 1e-8

        if not future:
            getLogger('probfoil').log(6, '%s: [BEST] x=%s (%s %s) -> %s' % (rule, max_x, tp_x, fp_x, score_x))
            rule.max_x = max_x
            rule.max_tp = max_tp
            rule.max_fp = max_fp
        return max_score


def main(argv=sys.argv[1:]):
    args = argparser().parse_args(argv)

    if args.seed:
        seed = args.seed
    else:
        seed = str(random.random())
    random.seed(seed)

    logger = 'probfoil'

    if args.log is None:
        logfile = None
    else:
        logfile = open(args.log, 'w')

    log = init_logger(verbose=args.verbose, name=logger, out=logfile)

    log.info('Random seed: %s' % seed)

    # Load input files
    data = DataFile(*(PrologFile(source) for source in args.files))

    if args.probfoil1:
        learn_class = ProbFOIL
    else:
        learn_class = ProbFOIL2

    time_start = time.time()
    learn = learn_class(data, logger=logger, **vars(args))

    hypothesis = learn.learn()
    time_total = time.time() - time_start

    print ('================ SETTINGS ================')
    for kv in vars(args).items():
        print('%20s:\t%s' % kv)

    if learn.interrupted:
        print('================ PARTIAL THEORY ================')
    else:
        print('================= FINAL THEORY =================')
    rule = hypothesis
    rules = rule.to_clauses(rule.target.functor)

    # First rule is failing rule: don't print it if there are other rules.
    if len(rules) > 1:
        for rule in rules[1:]:
            print (rule)
    else:
        print (rules[0])
    print ('==================== SCORES ====================')
    print ('            Accuracy:\t', accuracy(hypothesis))
    print ('           Precision:\t', precision(hypothesis))
    print ('              Recall:\t', recall(hypothesis))
    print ('================== STATISTICS ==================')
    for name, value in learn.statistics():
        print ('%20s:\t%s' % (name, value))
    print ('          Total time:\t%.4fs' % time_total)

    if logfile:
        logfile.close()


def argparser():
    parser = argparse.ArgumentParser()
    parser.add_argument('files', nargs='+')
    parser.add_argument('-1', '--det-rules', action='store_true', dest='probfoil1',
                        help='learn deterministic rules')
    parser.add_argument('-m', help='parameter m for m-estimate', type=float,
                        default=argparse.SUPPRESS)
    parser.add_argument('-b', '--beam-size', type=int, default=5,
                        help='size of beam for beam search')
    parser.add_argument('-p', '--significance', type=float, default=None,
                        help='rule significance threshold', dest='p')
    parser.add_argument('-l', '--length', dest='l', type=int, default=None,
                        help='maximum rule length')
    parser.add_argument('-v', action='count', dest='verbose', default=None,
                        help='increase verbosity (repeat for more)')
    parser.add_argument('--symmetry-breaking', action='store_true',
                        help='avoid symmetries in refinement operator')
    parser.add_argument('--target', '-t', type=str,
                        help='specify predicate/arity to learn (overrides settings file)')
    parser.add_argument('-s', '--seed', help='random seed', default=None)
    parser.add_argument('--log', help='write log to file', default=None)
    return parser


if __name__ == '__main__':
    main()


"""
Module name
"""

from __future__ import print_function

import math

def rates(rule):
    tp = 0.0
    fp = 0.0
    p = 0.0
    m = 0
    for c, pr in zip(rule.correct, rule.scores):
        tp += min(c, pr)
        fp += max(0, pr - c)
        p += c
        m += 1
    n = m - p
    return tp, fp, n - fp, p - tp


def m_estimate(rule, m=1):
    """Compute the m-estimate of the rule.

    :param rule: rule to score
    :param m: m parameter for m-estimate
    :return: value of m-estimate
    """
    tp, fp, tn, fn = rates(rule)
    p = tp + fn
    n = tn + fp
    return (tp + (m * p / (p + n))) / (tp + fp + m)


def m_estimate_future(rule, m=1):
    """Compute the m-estimate for the optimal extension of this rule.

    The optimal extension has the same TP-rate and zero FP-rate.

    :param rule: rule to score
    :param m: m parameter for m-estimate
    :return: value of m-estimate assuming an optimal extension
    """

    tp, fp, tn, fn = rates(rule)
    p = tp + fn
    n = tn + fp
    fp = 0.0
    return (tp + (m * p / (p + n))) / (tp + fp + m)


def m_estimate_relative(rule, m=1):
    """Compute the m-estimate of the rule relative to the previous ruleset.

    :param rule: rule to score
    :param m: m parameter for m-estimate
    :return: value of m-estimate
    """

    if rule.previous:
        tp_p, fp_p, tn_p, fn_p = rates(rule.previous)
    else:
        tp_p, fp_p, tn_p, fn_p = 0.0, 0.0, 0.0, 0.0   # last two are irrelevant

    tp, fp, tn, fn = rates(rule)
    p = tp + fn - tp_p
    n = tn + fp - fp_p
    return (tp - tp_p + (m * p / (p + n))) / (tp + fp - tp_p - fp_p + m)


def m_estimate_future_relative(rule, m=1):
    """Compute the m-estimate for the optimal extension of this rule relative to the previous ruleset.

    The optimal extension has the same TP-rate and zero FP-rate.

    :param rule: rule to score
    :param m: m parameter for m-estimate
    :return: value of m-estimate assuming an optimal extension
    """

    if rule.previous:
        tp_p, fp_p, tn_p, fn_p = rates(rule.previous)
    else:
        tp_p, fp_p, tn_p, fn_p = 0.0, 0.0, 0.0, 0.0  # last two are irrelevant

    tp, fp, tn, fn = rates(rule)
    p = tp + fn - tp_p
    n = tn + fp - fp_p
    fp = fp_p
    return (tp - tp_p + (m * p / (p + n))) / (tp + fp - tp_p - fp_p + m)


def accuracy(rule):
    tp, fp, tn, fn = rates(rule)
    return (tp + tn) / (tp + fp + tn + fn)


def precision(rule):
    tp, fp, tn, fn = rates(rule)
    if tp + fp == 0:
        return 0.0
    else:
        return tp / (tp + fp)


def recall(rule):
    tp, fp, tn, fn = rates(rule)
    return tp / (tp + fn)


def chi2_cdf(x):
    return math.erf(math.sqrt(x / 2))


def pvalue2chisquare(s, low=0.0, high=100.0, precision=1e-8):
    """Helper function for transforming significance p-value into ChiSquare decision value."""
    v = (low + high) / 2
    r = chi2_cdf(v)
    if -precision < r - s < precision:
        return v
    elif r > s:
        return pvalue2chisquare(s, low, v)
    else:
        return pvalue2chisquare(s, v, high)


def significance(rule, calc_max=False):
    """Compute the significance of a rule (chi-square distributed)."""

    c_tp, c_fp, c_tn, c_fn = rates(rule)

    pos = c_tp + c_fn
    neg = c_fp + c_tn

    if rule.previous:
        p_tp, p_fp, p_tn, p_fn = rates(rule.previous)
    else:
        p_tp, p_fp, p_tn, p_fn = 0.0, 0.0, neg, pos

    if calc_max:
        s_tp_max = c_tp     # TODO
        s_tp = s_tp_max - p_tp
        s_fp = 0
    else:
        s_tp = c_tp - p_tp
        s_fp = c_fp - p_fp

    s_pos = c_tp + c_fn
    s_neg = c_fp + c_tn
    s_all = s_pos + s_neg

    c = s_tp + s_fp     # max: c = s_tp
    if c == 0:
        return 0

    f_pos_c = s_tp / c     # max: f_pos_c = 1
    f_neg_c = 1 - f_pos_c  # max: f_neg_c == 0

    f_pos = s_pos / s_all
    f_neg = s_neg / s_all

    pos_log = math.log(f_pos_c / f_pos) if f_pos_c > 0 else 0  # max: pos_log = -log(sP / sM)
    neg_log = math.log(f_neg_c / f_neg) if f_neg_c > 0 else 0  # max: neg_log = 0

    l = 2 * c * (f_pos_c * pos_log + f_neg_c * neg_log)  # max: 2 * sTP * -log(sP/sM)

    return l




from problog.engine import DefaultEngine
from problog.logic import Term
from problog import get_evaluatable


class DataFile(object):
    """Represents a data file. This is a wrapper around a ProbLog file that offers direct
    querying and evaluation functionality.

    :param source: ProbLog logic program
    :type source: LogicProgram
    """

    def __init__(self, *sources):
        self._database = DefaultEngine().prepare(sources[0])
        for source in sources[1:]:
            for clause in source:
                self._database += clause

    def query(self, functor, arity=None, arguments=None):
        """Perform a query on the data.
        Either arity or arguments have to be provided.

        :param functor: functor of the query
        :param arity: number of arguments
        :type arity: int | None
        :param arguments: arguments
        :type arguments: list[Term]
        :return: list of argument tuples
        :rtype: list[tuple[Term]]
        """
        if arguments is None:
            assert arity is not None
            arguments = [None] * arity

        query = Term(functor, *arguments)
        return self._database.engine.query(self._database, query)

    def ground(self, rule, functor=None, arguments=None):
        """Generate ground program for the given rule.

        :param rule: rule to evaluate
        :type rule: Rule
        :param functor: override rule functor
        :type functor: str
        :param arguments: query arguments (None if non-ground)
        :type arguments: list[tuple[Term]] | None
        :return: ground program
        :rtype: LogicFormula
        """
        if rule is None:
            db = self._database
            target = Term(functor)
        else:
            db = self._database.extend()
            target = None
            for clause in rule.to_clauses(functor):
                target = clause.head
                db += clause

        if arguments is not None:
            queries = [target.with_args(*args) for args in arguments]
        else:
            queries = [target]

        return self._database.engine.ground_all(db, queries=queries)

    def evaluate(self, rule, functor=None, arguments=None, ground_program=None):
        """Evaluate the given rule.

        :param rule: rule to evaluate
        :type rule: Rule
        :param functor: override rule functor
        :type functor: str
        :param arguments: query arguments (None if non-ground)
        :type arguments: list[tuple[Term]] | None
        :param ground_program: use pre-existing ground program (perform ground if None)
        :type ground_program: LogicFormula | None
        :return: dictionary of results
        :rtype: dict[Term, float]
        """
        if ground_program is None:
            ground_program = self.ground(rule, functor, arguments)

        knowledge = get_evaluatable().create_from(ground_program)
        return knowledge.evaluate()



#! /usr/bin/env python

from __future__ import print_function
import sys
import os


version_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'probfoil', 'version.py')

version = {}
with open(version_file) as fp:
    exec(fp.read(), version)
version = version['version']

if __name__ == '__main__':
    from setuptools import setup, find_packages

    package_data = {}

    setup(
        name='probfoil',
        version=version,
        description='Prob2FOIL: rule learner for probabilistic logic',
        url='https://dtai.cs.kuleuven.be/software/probfoil',
        author='Anton Dries',
        author_email='anton.dries@cs.kuleuven.be',
        license='Apache Software License',
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: Apache Software License',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Prolog',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
        ],
        keywords='probabilistic logic learning',
        packages=find_packages(),
        entry_points={
            'console_scripts': ['probfoil=probfoil.probfoil:main']
        },
        package_data=package_data,
        install_requires=['problog']
    )


def increment_release(v):
    v = v.split('.')
    if len(v) == 4:
        v = v[:3] + [str(int(v[3]) + 1)]
    else:
        v = v[:4]
    return '.'.join(v)


def increment_dev(v):
    v = v.split('.')
    if len(v) == 4:
        v = v[:3] + [str(int(v[3]) + 1), 'dev1']
    else:
        v = v[:4] + ['dev' + str(int(v[4][3:]) + 1)]
    return '.'.join(v)


def increment_version_dev():
    v = increment_dev(version)
    os.path.dirname(__file__)
    with open(version_file, 'w') as f:
        f.write("version = '%s'\n" % v)


def increment_version_release():
    v = increment_release(version)
    with open(version_file, 'w') as f:
        f.write("version = '%s'\n" % v)



from __future__ import print_function

from problog.logic import Var, Term, Constant
from itertools import product
from problog.util import Timer
import logging


class KnownError(Exception):
    pass


class LearnEntail(object):

    def __init__(self, data, language, target=None, logger=None, **kwargs):
        self._language = language

        if target is not None:
            try:
                t_func, t_arity = target.split('/')
                target = Term(t_func, int(t_arity))
            except Exception:
                raise KnownError('Invalid target specification \'%s\'' % target)

        self._target = target
        self._examples = None
        self._logger = logger

        self._data = data
        self._scores_correct = None

    @property
    def target(self):
        """The target predicate of the learning problem."""
        return self._target

    @property
    def examples(self):
        """The examples (tuple of target arguments) of the learning problem."""
        return self._examples

    @property
    def language(self):
        """Language specification of the learning problem."""
        return self._language

    def load(self, data):
        """Load the settings from a data file.

        Initializes language, target and examples.

        :param data: data file
        :type data: DataFile
        """
        self.language.load(data)  # for types and modes

        if self._target is None:
            try:
                target = data.query('learn', 1)[0]
                target_functor, target_arity = target[0].args
            except IndexError:
                raise KnownError('No target specification found!')
        else:
            target_functor, target_arity = self._target.functor, self._target.args[0]
        target_arguments = [Var(chr(65 + i)) for i in range(0, int(target_arity))]
        self._target = Term(str(target_functor), *target_arguments)

        # Find examples:
        #  if example_mode is closed, we will only use examples that are defined in the data
        #      this includes facts labeled with probability 0.0 (i.e. negative example)
        #  otherwise, the examples will consist of all combinations of values appearing in the data
        #      (taking into account type information)
        example_mode = data.query(Term('example_mode'), 1)
        if example_mode and str(example_mode[0][0]) == 'auto':
            types = self.language.get_argument_types(self._target.functor, self._target.arity)
            values = [self.language.get_type_values(t) for t in types]
            self._examples = list(product(*values))
        elif example_mode and str(example_mode[0][0]) == 'balance':
            # Balancing based on count only
            pos_examples = [r for r in data.query(self._target.functor, self._target.arity)]
            pos_count = len(pos_examples)
            types = self.language.get_argument_types(self._target.functor, self._target.arity)
            values = [self.language.get_type_values(t) for t in types]
            from random import shuffle
            neg_examples = list(product(*values))
            shuffle(neg_examples)
            logger = logging.getLogger(self._logger)
            logger.debug('Generated negative examples:')
            for ex in neg_examples[:pos_count]:
                logger.debug(Term(self._target(*ex).with_probability(0.0)))

            self._examples = pos_examples + neg_examples[:pos_count]
        else:
            self._examples = [r for r in data.query(self._target.functor, self._target.arity)]

        with Timer('Computing scores', logger=self._logger):
            self._scores_correct = self._compute_scores_correct()

    def _compute_scores_correct(self):
        """Computes the score for each example."""
        result = self._data.evaluate(rule=None, functor=self.target.functor, arguments=self.examples)

        scores_correct = []
        for example in self.examples:
            scores_correct.append(result.get(Term(self.target.functor, *example), 0.0))
        return scores_correct

    def _compute_scores_predict(self, rule):
        return self._compute_scores_predict_ground(rule)

    def _compute_scores_predict_ground(self, rule):
        functor = 'eval_rule'

        # Don't evaluate examples that are guaranteed to evaluate to 0.0 or 1.0.
        set_one = []
        set_zero = []
        if rule.previous is not None:
            set_one = [i for i, s in enumerate(rule.previous.scores) if s > 1 - 1e-8]
        if rule.parent is not None:
            set_zero = [i for i, s in enumerate(rule.parent.scores) if s < 1e-8]

        to_eval = set(range(0, len(self.examples))) - set(set_one) - set(set_zero)
        examples = [self.examples[i] for i in to_eval]
        # print (len(set_zero), len(set_one))

        # message = ''
        # for ex in examples:
        #     message += str(Term('query', Term(functor, *ex))) + '.\n'
        # message += '\n'.join(map(str, rule.to_clauses(functor))) + '\n'

        # Call ProbLog
        result = self._data.evaluate(rule, functor=functor, arguments=examples)

        # Extract results
        scores_predict = [0.0] * len(self.examples)
        for i, example in zip(to_eval, examples):
            scores_predict[i] = result.get(Term(functor, *example), 0.0)
        for i in set_one:
            scores_predict[i] = 1.0

        return scores_predict

    # def _compute_scores_predict_nonground(self, rule):
    #     """Evaluate the current rule using a non-ground query.
    #
    #       This is not possible because we can't properly distribute the weight of the
    #        non-ground query over the possible groundings.
    #       So this only works when all rules in the ruleset are range-restricted.
    #
    #     :param rule:
    #     :return:
    #     """
    #     functor = 'eval_rule'
    #     result = self._data.evaluate(rule, functor=functor, arguments=[self._target.args])
    #
    #     types = None
    #     values = None
    #
    #     from collections import defaultdict
    #     from problog.logic import is_variable
    #
    #     index = defaultdict(dict)
    #     for key, value in result.items():
    #         if not key.is_ground():
    #             if values is None:
    #                 types = self.language.get_argument_types(self._target.functor, self._target.arity)
    #                 values = [len(self.language.get_type_values(t)) for t in types]
    #             c = 1
    #
    #             gi = []
    #             gk = []
    #             for i, arg, vals in zip(range(0, len(key.arity)), key.args, values):
    #                 if is_variable(key.args):
    #                     c *= vals
    #                 else:
    #                     gi.append(i)
    #                     gk.append(arg)
    #             import math
    #             p = 1.0 - value ** (1.0 / c)
    #             p = 1 - math.exp(math.log(1 - value) / c)
    #             index[tuple(gi)][tuple(gk)] = [p, c]
    #
    #     scores_predict = [0.0]
    #     for i, arg in enumerate(self.examples):
    #         for gi, idx in index.items():
    #             gk = tuple(arg[j] for j in gi)
    #             res = idx.get(gk, [0.0, 0])
    #             scores_predict
    #
    #     print (rule, result)
    #     print (self.examples)


class CandidateSet(object):

    def __init__(self):
        pass

    def push(self, candidate):
        raise NotImplementedError('abstract method')

    def pop(self):
        raise NotImplementedError('abstract method')

    def __bool__(self):
        raise NotImplementedError('abstract method')


class BestCandidate(CandidateSet):

    def __init__(self, candidate=None):
        CandidateSet.__init__(self)
        self.candidate = candidate

    def push(self, candidate):
        if self.candidate is None or self.candidate.score_cmp < candidate.score_cmp:
            self.candidate = candidate

    def pop(self):
        if self.candidate is not None:
            return self.candidate
        else:
            raise IndexError('Candidate set is empty!')

    def __bool__(self):
        return not self.candidate is None


class CandidateBeam(CandidateSet):

    def __init__(self, size):
        CandidateSet.__init__(self)
        self._size = size
        self._candidates = []

    def _bottom_score(self):
        if self._candidates:
            return self._candidates[-1].score_cmp
        else:
            return -1e1000

    def _insert(self, candidate):
        for i, x in enumerate(self._candidates):
            if x.is_equivalent(candidate):
                raise ValueError('duplicate')
            elif x.score_cmp < candidate.score_cmp:
                self._candidates.insert(i, candidate)
                return False
        self._candidates.append(candidate)
        return True

    def push(self, candidate):
        """Adds a candidate to the beam.

        :param candidate: candidate to add
        :return: True if candidate was accepted, False otherwise
        """
        if len(self._candidates) < self._size or candidate.score_cmp > self._bottom_score():
            #  We should add it to the beam.
            try:
                is_last = self._insert(candidate)
                if len(self._candidates) > self._size:
                    self._candidates.pop(-1)
                    return not is_last
            except ValueError:
                return False
            return True
        return False

    def pop(self):
        return self._candidates.pop(0)

    def __bool__(self):
        return bool(self._candidates)

    def __nonzero__(self):
        return bool(self._candidates)

    def __str__(self):
        s = '==================================\n'
        for candidate in self._candidates:
            s += str(candidate) + ' ' + str(candidate.score) + '\n'
        s += '=================================='
        return s
